package service

import (
	"context"

	"git.nobla.cn/golang/aeus-admin/models"
	"git.nobla.cn/golang/aeus-admin/pb"
	"git.nobla.cn/golang/aeus-admin/types"
	"git.nobla.cn/golang/aeus/middleware/auth"
	"git.nobla.cn/golang/aeus/pkg/errors"
	"gorm.io/gorm"
)

type (
	userOptions struct {
		db *gorm.DB
	}
	UserOption func(o *userOptions)

	UserService struct {
		opts *userOptions
	}
)

func WithUserDB(db *gorm.DB) UserOption {
	return func(o *userOptions) {
		o.db = db
	}
}

func (s *UserService) getUidFromContext(ctx context.Context) (string, error) {
	if claims, ok := auth.FromContext(ctx); !ok {
		return "", errors.ErrAccessDenied
	} else {
		return claims.GetSubject()
	}
}

func (s *UserService) hasPermission(menuID int64, permissions []*models.Permission) bool {
	for _, permission := range permissions {
		if permission.MenuId == menuID {
			return true
		}
	}
	return false
}

func (s *UserService) getPermissions(menuID int64, permissions []*models.Permission) []*pb.PermissionItem {
	ss := make([]*pb.PermissionItem, 0, 10)
	for _, permission := range permissions {
		if permission.MenuId == menuID {
			ss = append(ss, &pb.PermissionItem{
				Value: permission.Permission,
				Label: permission.Label,
			})
		}
	}
	return ss
}

func (s *UserService) recursiveNestedMenu(ctx context.Context, parent int64, perm bool, menus []*models.Menu, permissions []*models.Permission) []*pb.MenuItem {
	values := make([]*pb.MenuItem, 0)
	for _, row := range menus {
		if row.ParentId == parent {
			if !row.Public && !s.hasPermission(row.Id, permissions) {
				continue
			}
			v := &pb.MenuItem{
				Label:    row.Label,
				Name:     row.Name,
				Icon:     row.Icon,
				Hidden:   row.Hidden,
				Route:    row.Uri,
				Public:   row.Public,
				View:     row.ViewPath,
				Children: s.recursiveNestedMenu(ctx, row.Id, perm, menus, permissions),
			}
			if perm {
				v.Permissions = s.getPermissions(row.Id, permissions)
			}
			values = append(values, v)
		}
	}
	return values
}

func (s *UserService) getRolePermissions(ctx context.Context, db *gorm.DB, role string) (permissions []*models.Permission, err error) {
	permissions = make([]*models.Permission, 0)
	err = db.Where("id IN (SELECT permission_id FROM role_permissions WHERE role = ?)", role).Find(&permissions).Error
	return
}

func (s *UserService) GetMenus(ctx context.Context, req *pb.GetMenuRequest) (res *pb.GetMenuResponse, err error) {
	claims, ok := auth.FromContext(ctx)
	if !ok {
		return nil, errors.ErrAccessDenied
	}
	var (
		permissions []*models.Permission
	)
	tx := s.opts.db.WithContext(ctx)
	if claims, ok := claims.(*types.Claims); ok {
		permissions, err = s.GetRolePermissions(ctx, tx, claims.Role)
	}
	values := make([]*models.Menu, 0)
	if err = tx.Find(&values).Error; err != nil {
		return
	}
	res = &pb.GetMenuResponse{
		Data: s.recursiveNestedMenu(ctx, 0, req.Permission, values, permissions),
	}
	return
}

func (s *UserService) GetProfile(ctx context.Context, req *pb.GetProfileRequest) (res *pb.GetProfileResponse, err error) {
	if req.Uid == "" {
		if req.Uid, err = s.getUidFromContext(ctx); err != nil {
			return
		}
	}
	res = &pb.GetProfileResponse{}
	tx := s.opts.db.WithContext(ctx)
	err = tx.Table("users AS u").
		Select("u.uid as uid", "u.username", "u.avatar", "u.email", "u.description", "u.role", "u.admin").
		Where("u.uid=? ", req.Uid).
		First(res).Error
	return
}

func (s *UserService) UpdateProfile(ctx context.Context, req *pb.UpdateProfileRequest) (res *pb.UpdateProfileResponse, err error) {
	if req.Uid == "" {
		if req.Uid, err = s.getUidFromContext(ctx); err != nil {
			return
		}
	}
	userModel := &models.User{}
	tx := s.opts.db.WithContext(ctx)
	if err = tx.Where("uid=?", req.Uid).First(userModel).Error; err != nil {
		return
	}
	if req.Avatar != "" {
		userModel.Avatar = req.Avatar
	}
	if req.Email != "" {
		userModel.Email = req.Email
	}
	if req.Username != "" {
		userModel.Username = req.Username
	}
	if req.Description != "" {
		userModel.Description = req.Description
	}
	if err = tx.Save(userModel).Error; err == nil {
		res = &pb.UpdateProfileResponse{
			Uid: userModel.Uid,
		}
	}
	return
}

func (s *UserService) ResetPassword(ctx context.Context, req *pb.ResetPasswordRequest) (res *pb.ResetPasswordResponse, err error) {
	if req.Uid == "" {
		if req.Uid, err = s.getUidFromContext(ctx); err != nil {
			return
		}
	}
	userModel := &models.User{}
	tx := s.opts.db.WithContext(ctx)
	if err = tx.Where("uid=?", req.Uid).First(userModel).Error; err != nil {
		return
	}
	if userModel.Password == req.OldPassword {
		if err = tx.Where("uid=?", req.Uid).Model(&models.User{}).UpdateColumn("password", req.NewPassword).Error; err == nil {
			res = &pb.ResetPasswordResponse{
				Uid: userModel.Uid,
			}
		}
	} else {
		err = errors.Format(errors.AccessDenied, "invalid old password")
	}
	return
}

func (s *UserService) GetPermissions(ctx context.Context, req *pb.GetPermissionRequest) (res *pb.GetPermissionResponse, err error) {
	if req.Uid == "" {
		if req.Uid, err = s.getUidFromContext(ctx); err != nil {
			return
		}
	}
	userModel := &models.User{}
	tx := s.opts.db.WithContext(ctx)
	if err = userModel.FindOne(tx, "uid=?", req.Uid); err != nil {
		return
	}
	var permissions []string
	tx.Select("permission").Where("id IN (SELECT permission_id FROM role_permissions WHERE role =?)", userModel.Role).Model(&models.Permission{}).Pluck("permission", &permissions)
	res = &pb.GetPermissionResponse{
		Uid:         userModel.Uid,
		Permissions: permissions,
	}
	return
}

func (s *UserService) GetUserLabels(ctx context.Context, req *pb.GetUserLabelRequest) (res *pb.GetUserLabelResponse, err error) {
	values := make([]*models.User, 0)
	if err = s.opts.db.WithContext(ctx).Find(&values).Error; err == nil {
		res = &pb.GetUserLabelResponse{
			Data: make([]*pb.LabelValue, 0, len(values)),
		}
		for _, v := range values {
			res.Data = append(res.Data, &pb.LabelValue{
				Label: v.Username,
				Value: v.Uid,
			})
		}
	}
	return
}

func (s *UserService) GetUserTags(ctx context.Context, req *pb.GetUserTagRequest) (res *pb.GetUserTagResponse, err error) {
	values := make([]*models.User, 0)
	if err = s.opts.db.WithContext(ctx).Select("DISTINCT(`tag`) AS `tag`").Find(&values).Error; err == nil {
		res = &pb.GetUserTagResponse{
			Data: make([]*pb.LabelValue, 0, len(values)),
		}
		for _, v := range values {
			res.Data = append(res.Data, &pb.LabelValue{
				Label: v.Tag,
				Value: v.Tag,
			})
		}
	}
	return
}

func NewUserService(cbs ...UserOption) *UserService {
	opts := &userOptions{}
	for _, cb := range cbs {
		cb(opts)
	}
	return &UserService{
		opts: opts,
	}
}
